#ifndef __POLYRAWWRITER_H
#define __POLYRAWWRITER_H

//polyRawWriter.h
//
// *****************************************************************************
//
// Copyright (C) 1997-2002 Alias|Wavefront, a division of Silicon Graphics Limited.
// 
// The information in this file is provided for the exclusive use of the
// licensees of Alias|Wavefront.  Such users have the right to use, modify,
// and incorporate this code into other products for purposes authorized
// by the Alias|Wavefront license agreement, without fee.
// 
// ALIAS|WAVEFRONT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
// EVENT SHALL ALIAS|WAVEFRONT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
// CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
// DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
// TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
// PERFORMANCE OF THIS SOFTWARE.
//
// *****************************************************************************
//
// CLASS:    polyRawWriter
//
// *****************************************************************************
//
// CLASS DESCRIPTION (polyRawWriter)
// 
// polyRawWriter is a class derived from polyWriter.  It currently outputs in
// raw text format the following polygonal mesh data:
// - faces and their vertex components
// - vertex coordinates
// - colors per vertex
// - normals per vertex
// - current uv set and coordinates
// - component sets
// - file textures (for the current uv set)
// - other uv sets and coordinates
//
// *****************************************************************************

#include <maya/MFnMesh.h>
#include <maya/MPointArray.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MFloatArray.h>



//Used to store UV set information
//
struct UVSet {
	MFloatArray	uArray;
	MFloatArray	vArray;
	MString		name;
	UVSet*		next;
};

class polyRawWriter  {

	public:
				polyRawWriter (const MDagPath& dagPath, MStatus& status);
		virtual		~polyRawWriter ();
				MStatus extractGeometry ();
				MStatus writeShape (ostream& os);
		    
				void	doTabs(ostream& os);
				int	tabs;





	private:
		//Functions
		//
		MStatus	outputSingleSet (ostream& os, 
						 MString setName, 
						 MIntArray faces, 
						 MString textureName);
		MStatus outputFaces (ostream& os);
		MStatus outputVertices (ostream& os);
		MStatus	outputVertexInfo (ostream& os);
		MStatus	outputNormals (ostream& os);
		MStatus	outputColors (ostream& os);
		MStatus	outputUVs (ostream& os);
			
		MObject jointForCluster(MObject& jointCluster) const;
		MStatus outputJoints(ostream& os) ;
		MStatus outputJointCluster(ostream& os); 


		MObject	findShader (const MObject& setNode);
		virtual MStatus	outputSets (ostream& os);
		
	
		//Data Member
		//the current UV set's name
			    
		MString			fCurrentUVSetName;

		//for storing general mesh information
		//
		MPointArray		fVertexArray;
		MColorArray		fColorArray;
		MFloatVectorArray	fNormalArray;

		//for storing DAG objects
		//
		MFnMesh*		fMesh;
		MDagPath*		fDagPath;
		MObjectArray		fPolygonSets;
		MObjectArray		fPolygonComponents;
		//for storing UV information
		UVSet*			fHeadUVSet;
};

#endif /*__POLYRAWWRITER_H*/

